import math, geometry, random, time, os
import numpy, pylab

def format_time(seconds):
    if seconds<60:
        return "%d seconds" % seconds
    if seconds>60 and seconds<3600:
        return "%d minutes %d seconds" % (seconds//60, seconds%60)  
    if seconds>=3600:
        return "%d hours %d minutes %d seconds" % (seconds//3600, (seconds//60)%60, seconds%60)
         

class NavStats:
    def __init__(self):
        self.log = []
        
    def update(self, wanderers, centroid_pos, max_distance, elapsed, converged):
        # record the statistics that change
        row = {}
        row['elapsed'] = elapsed
        row['centroid'] = centroid_pos
        row['max_distance'] = max_distance
        row['converged'] = converged
        
        log_wanderers = []
        for wanderer in wanderers:
            
            subrow = {}
            subrow['pos'] = wanderer.pos
            subrow['offset_pos'] = wanderer.offset_pos
            subrow['halting'] = wanderer.halting
            subrow['heading'] = wanderer.heading
            subrow['offset_angle'] = wanderer.offset_angle
            subrow['distance'] = wanderer.distance
            subrow['true_target_distance'] = wanderer.true_target_distance
            log_wanderers.append(subrow)
        row['wanderers'] = log_wanderers
        self.log.append(row)
        
         
                
def format_tex_report(world, logs, iterations, stats):
    texheader = """
\\documentclass{article}

\\usepackage{epsfig}
\\usepackage{times}

\\begin{document}    
\section*{Rendezvous simulator log report %s}
\\setlength{\\parindent}{0pt} 
\\setcounter{secnumdepth}{0}
\\pagestyle{empty}

\\renewcommand{\\topfraction}{0.9}
\\renewcommand{\\textfraction}{0.1}




\\newcommand{\\cnfig}[3]{\\begin{figure}[htbp]\\begin{center}
	\\epsfig{file = #1.eps, width = #3}
\\caption{#2}\label{fig:#1}
\\end{center}\\end{figure}\\ }
    """ % (time.asctime())
    
    
    texparameters_page = """
    \\subsection{Global Parameters}
    Number of agents: %(agents)d\\\\
    Number of iterations: %(iterations)d\\\\
    Field size: %(field_size)sm $\\times$  %(field_size).0fm\\\\
    Convergence radius: %(radius).1fm\\\\
    \\subsection{Agent Parameters}    
    GPS position uncertainty (Gaussian noise, std. dev.): %(gps).1fm\\\\
    Heading sensor uncertainty (Gaussian noise, std. dev.): %(angle).2f$^\\circ$\\\\
    Target acquisition mode: %(mode)s\\\\
    Angle update time (avg., exp. distribution): %(ang_time).1fs\\\\
    GPS noise update time (avg., exp. distribution): %(gps_time).1fs\\\\
    Walking speed: %(speed)s m/s\\\\
    Pausing time to acquire new heading: %(halt).0fs\\\\
    \\subsection{Centroid parameters}
    """ % {
    "agents" : world.n_wanderers,
    "iterations" : iterations,
    "field_size" : world.field_size,
    "radius" : world.stop_radius,
    "gps" : world.wanderer_args["position_uncertainty"],
    "angle" : world.wanderer_args["angle_uncertainty"],
    "mode" : world.wanderer_args["mode"],
    "ang_time" : 1.0/world.wanderer_args["angle_update_prob"],
    "gps_time" : 1.0/world.wanderer_args["offset_update_prob"],
    "speed" : world.wanderer_args["step_size"],
    "halt" : world.wanderer_args["halt_time"]
    }
    
    
    if world.sprung:
        tex_centroid = """
        Spring centroid mode\\\\
        Spring K: %(k).2f\\\\
        Spring damping: %(damping).4f\\\\
        Spring breaking %(breaking).1f\\\\
        """ % world.spring_args
    else:
        tex_centroid = """
        Simple geometric centroid mode\\\\"""
        
    
    tex_results = """\\hrule 
    \\subsection{Results}
    Mean time to convergence: %(mean_time)s (std. dev. %(std_time)s)\\\\
    Mean distance walked until convergence (per agent): %(mean_distance).0fm (std. dev. %(std_distance).0fm)\\\\
    Mean total turning per agent: %(mean_turning).1f$^\\circ$ (std. dev. %(std_turning).1f$^\\circ$)\\\\
    """%stats
    
    texparameters_page = texparameters_page + tex_centroid + tex_results+"\n \\pagebreak"
    
    
        
            
        
    figures = """\cnfig{convtime}{Histogram of end times when all agents reached target (i.e. were within %.1f meters of the centroid.)}{\\linewidth}\n""" % world.stop_radius
    
    figures += """\cnfig{convspeed}{Histogram of overall speed of acquisition. This is the total time to converge divided by the distance of initially furthest agent from the centroid.}{\\linewidth}\n""" 
    
    figures += """\cnfig{walktime}{Maximum distance of any agent from the centroid against time for each of the runs. Each run is a separate line.}{\\linewidth}\n""" 
    
    figures += """\cnfig{centroids}{Trajectories of centroids for a small sample of the runs. Each colour represents a different run. Trajectories begin at $\\circ$ and end at $\\times$. The blue dotted line represents the path of the centroid.}{\\linewidth}\n""" 
    
    
    figures += """\\begin{figure}
    
    \\epsfig{file=agents0.eps, width=0.49\\linewidth}
    \\epsfig{file=agents1.eps, width=0.49\\linewidth}
    \\epsfig{file=agents2.eps, width=0.49\\linewidth}
    \\epsfig{file=agents3.eps, width=0.49\\linewidth}
    \\epsfig{file=agents4.eps, width=0.49\\linewidth}
    \\epsfig{file=agents5.eps, width=0.49\\linewidth}
    \\caption{Trajectories of the agents for a small sample of the runs. Trajectories begin at $\\circ$ and end at $\\times$. Solid colors are true positions; dotted lines are the positions estimated by GPS. The thick line is the position of the centroid.}
    \\end{figure}""" 
    
    
   
    
        
    
    
    texfooter = """
\end{document}    
    """
    
    return texheader+texparameters_page+figures+texfooter
        
        
        
def create_stats(world, logs, iterations):
    stats = {}
    
    # compute average time to converge
    end_times = numpy.array([log[-1]['elapsed'] for log in logs])
    
    f = open("end_times.txt", "w")
    for end in end_times:
        f.write("%d\n" % end)
    f.close()
    
    
    stats["mean_time"] = format_time(numpy.mean(end_times))
    stats["std_time"] = format_time(numpy.std(end_times))
    
    # compute average distance walked
    distances = []    
    for i in range(world.n_wanderers):
        end_wanderers = numpy.array([log[-1]['wanderers'][i]['distance'] for log in logs])
        distances.append(end_wanderers)
        
        
    
    distances = numpy.hstack(distances)
    stats["mean_distance"] = numpy.mean(distances)
    stats["std_distance"] = numpy.std(distances)

    #compute total amount of turning
    angle_sums = []
    for log in logs:
        for i in range(world.n_wanderers):
            angles = []
            for entry in log:
                angles.append(entry['wanderers'][i]['heading'])
                
            diff = numpy.diff(numpy.array(angles))
            sum_angles = numpy.sum(numpy.abs(diff))
            angle_sums.append(sum_angles)
                    
                
    stats["mean_turning"] = numpy.mean(angle_sums)*(180.0/math.pi)
    stats["std_turning"] = numpy.std(angle_sums)*(180.0/math.pi)
        
    
    return stats
    
        
    
    
    

def create_report(world, logs, iterations, directory="report"):
    
    try:
        os.mkdir(directory)
    except:
        pass
    os.chdir(directory)
    
    
    stats = create_stats(world, logs, iterations)
    field_size = world.field_size
    axrange = [-field_size/1.5, field_size/1.5, -field_size/1.5, field_size/1.5]
    end_times = numpy.array([log[-1]['elapsed'] for log in logs])
    init_distances = numpy.array([log[0]['max_distance'] for log in logs])
    
    # raw times    
    pylab.figure(1)
    pylab.title("End times for convergence (mean=%.1fs, std=%.1fs)" % (numpy.mean(end_times), numpy.std(end_times)))
    pylab.ylabel("Frequency")
    pylab.xlabel("Time (s)")
    
    pylab.hist(end_times, 20)
    pylab.axis([0,3600,0,iterations/3])
    pylab.savefig("convtime.eps")
    
    # speed (max_distance/time)
    pylab.figure(2)
    speed = init_distances/(end_times+1)
    pylab.title("Overall Speed (mean=%.2f m/s, std=%.2f m/s)" % (numpy.mean(speed), numpy.std(speed)))
    pylab.ylabel("Frequency")
    pylab.xlabel("Speed (m/s)")        
    
    pylab.hist(speed, 20)
    pylab.axis([0,1.0,0,iterations/3])
    pylab.savefig("convspeed.eps")
    
    # distance to target
    pylab.figure(3)
    pylab.hold(True)
    for log in logs:
        max_distance = numpy.array([v['max_distance'] for v in log])
        pylab.plot(max_distance, 'b')
    pylab.title("Maximum distance of wanderer to target")
    pylab.ylabel("Distance (m)")
    pylab.xlabel("Time (s)")        
    pylab.savefig("walktime.eps")
    
    
    # centroid positions
    random.shuffle(logs)
    pylab.figure(4)
    pylab.hold(True)
    
    # obstacles
    for obstacle in world.obstacles:
        pylab.fill([pt[0] for pt in obstacle.pts],[pt[1] for pt in obstacle.pts], facecolor="#d0d0d0")
        
    
    
    for log in logs[0:6]:        
        positions = numpy.array([v['centroid'] for v in log])
        
        pylab.plot(positions[:,0], positions[:,1])
        pylab.plot([positions[0,0]], [positions[0,1]], 'o')
        pylab.plot([positions[-1,0]], [positions[-1,1]], 'x')
        
    pylab.axis(axrange)
    pylab.title("Trajectories of centroid for 6 example runs.")
    pylab.ylabel("Y position (m)")
    pylab.xlabel("X position (m)")        
    pylab.savefig("centroids.eps")
    
    
    
    # agent positions
    
    pylab.figure(5)
    pylab.hold(False)
       
    colors = "rgbcmyk"
    k = 0
    for log in logs[0:6]:                
        centroid = numpy.array([v['centroid'] for v in log])
        pylab.clf()
            
            
        # centroid
        pylab.hold(True)
        pylab.plot(centroid[:,0], centroid[:,1], 'bo')
        pylab.plot([centroid[0,0]], [centroid[0,1]], 'bo')
        pylab.plot([centroid[-1,0]], [centroid[-1,1]], 'bx')
        
        
        # obstacles
        for obstacle in world.obstacles:
            pylab.fill([pt[0] for pt in obstacle.pts],[pt[1] for pt in obstacle.pts], facecolor="#d0d0d0")
        
        # agents
        
        for i in range(world.n_wanderers):            
            positions = numpy.array([v['wanderers'][i]['pos'] for v in log])
            opositions = numpy.array([v['wanderers'][i]['offset_pos'] for v in log])
            
            c = colors[i%len(colors)]
            pylab.plot(opositions[:,0], opositions[:,1], c+":")
            pylab.plot(positions[:,0], positions[:,1], c)
            
            
            pylab.plot([positions[0,0]], [positions[0,1]], 'o')
            pylab.plot([positions[-1,0]], [positions[-1,1]], 'x')
            
            
            pylab.axis(axrange)
            
            pylab.title("Trajectories of %d agents." % world.n_wanderers)
            pylab.ylabel("Y position (m)")
            pylab.xlabel("X position (m)")
            pylab.savefig("agents%d.eps" % k)            
    
            
            
            
        k = k + 1
        
    
    # agent quiver plots
    
    # pylab.figure(6)
    # pylab.hold(False)
        
    # k = 0
    # for log in logs[0:6]:                
        # centroid = numpy.array([v['centroid'] for v in log])
        # pylab.clf()
            
        # pylab.hold(True)
        # pylab.plot(centroid[:,0], centroid[:,1], 'b:')
        # pylab.plot([centroid[0,0]], [centroid[0,1]], 'bo')
        # pylab.plot([centroid[-1,0]], [centroid[-1,1]], 'bx')
        
            
        # for i in range(world.n_wanderers):            
            # positions = numpy.array([v['wanderers'][i]['pos'] for v in log])
            
            # headings = numpy.array([v['wanderers'][i]['heading'] for v in log])
            # u = numpy.cos(headings)
            # v = -numpy.sin(headings)
            
            # pylab.quiver(positions[:,0], positions[:,1], u, v)
            # pylab.plot([positions[0,0]], [positions[0,1]], 'o')
            # pylab.plot([positions[-1,0]], [positions[-1,1]], 'x')
            
            
            # pylab.axis(axrange)
            
            # pylab.title("Trajectories and headings of %d agents." % world.n_wanderers)
            # pylab.ylabel("Y position (m)")
            # pylab.xlabel("X position (m)")
            # pylab.savefig("agentsquiver%d.eps" % k)            
    
            
            
            
        # k = k + 1

    
    #create the latex file
    try:
        os.remove("report.tex")
    except:
        print "Couldn't delete.."
        pass
    

    for i in range(5):
        pylab.figure(i)
        pylab.clf()
        
    tex_file = open("report.tex", "w")    
    tex_file.write(format_tex_report(world, logs, iterations, stats))
    tex_file.close()
    
    # latex it and pdf it
    os.system("latex report")
    os.system("latex report")
    os.system("dvipdfm report")
    
    os.chdir("..")
    os.system("report\\report.pdf")
    
    
 